# Copyright (c) 2023 Johnathan P. Irvin and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
from .prompt import Prompt

FEATURE_REQUEST_PROMPT = Prompt(
    content="""
Fill out this form:

```
---
name: <Name Of Feature>
about: <Short Description>
labels: < Documentation, Enhancement, Question, Bug, etc. >
---

**Is your feature request related to a problem? Please describe.**
A clear and concise description of what the problem is. Ex. I'm always frustrated when [...]

**Describe the solution you'd like**
A clear and concise description of what you want to happen.

**Describe alternatives you've considered**
A clear and concise description of any alternative solutions or features you've considered.
```
""")

BUG_REPORT_PROMPT = Prompt(
    content="""
Fill out this form with as much detail as possible with the provided information:
```
---
name: <Name Of Bug>
about: <Short Description>
labels: < Documentation, Enhancement, Question, Bug, etc. >
---

**Describe the bug**
A clear and concise description of what the bug is.

**To Reproduce**
Steps to reproduce the behavior:
1. Go to '...'
2. Click on '....'
3. Scroll down to '....'
4. See error

**Expected behavior**
A clear and concise description of what you expected to happen.
""")



# Copyright (c) 2023 Johnathan P. Irvin and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
from dataclasses import dataclass, field

MODEL_MESSAGE = list[dict[str, str]]

@dataclass(frozen=True)
class Prompt:
    """
    A prompt is a guide for OpenAI models to generate text. It is a
    string that is used to seed the model's text generation. It is
    typically a question or a statement that the model is asked to
    respond to.

    Examples refer to "user" messages that are sent to the model, and
    "assistant" messages that are sent from the model. We use this
    as additional context for the model to generate text.
    """    
    content: str
    examples: list[MODEL_MESSAGE] = field(default_factory=list)

    @property
    def messages(self) -> list[MODEL_MESSAGE]:
        """
        Get the messages to send to the model. This includes the prompt
        and any examples. The prompt is sent as a "system" message, and
        the examples are sent as "user" and "assistant" messages.

        The dictionary is formatted as follows:
        {
            "role": <"system" | "user" | "assistant">,
            "content": <str>
        }

        Returns:
            list[MODEL_MESSAGE]: The messages to send to the model.
        """        
        return [
            {
                "role": "user",
                "content": self.content
            }
        ] + self.examples



# Copyright (c) 2023 Johnathan P. Irvin and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
from argparse import ArgumentParser, Namespace
from importlib import import_module
from os import listdir
from os.path import dirname
from sys import argv, stdin


def _has_required_args(module: "ModuleType", args: Namespace) -> bool:
    """
    Checks if the module has the required arguments. If the module does not
    have the ARGUMENTS attribute, then it is assumed that the module does not
    require any arguments. If the module does have the ARGUMENTS attribute,
    then it is assumed that the module requires arguments as long as the
    ARGUMENTS attribute is a dictionary
    
    If the ARGUMENTS attribute is a dictionary, then the module will be checked
    to see if it has all of the required arguments. If the module has all of the
    required arguments, then the module will be checked to see if the values of
    the required arguments match the values of the arguments passed to the parser.
    
    If the values of the required arguments match the values of the arguments
    passed to the parser, are in a list of acceptable values, or pass a callable
    function, then the module will be considered to have the required arguments.

    Args:
        module (ModuleType): The module to check for required arguments.
        args (Namespace): The arguments passed to the parser.

    Returns:
        bool: True if the module has the required arguments, False otherwise.
    """    
    if not hasattr(module, "ARGUMENTS"):
        return True
    
    if not isinstance(module.ARGUMENTS, dict):
        return True
    
    for key in module.ARGUMENTS:
        if key not in args:
            return False
        
        arg_value = getattr(args, key)
        mod_value = module.ARGUMENTS[key]

        if isinstance(mod_value, str) and arg_value != mod_value:
            return False

        if isinstance(mod_value, list) and arg_value not in mod_value:
            return False
        
        if callable(mod_value) and not mod_value(arg_value):
            return False
            
    return True

def _add_args_parser(parser: ArgumentParser, module: "ModuleType", args: Namespace):
    """
    Adds the given module's subparsers and arguments to the parser.
    
    Args:
        parser (ArgumentParser): The parser to add the module to.
        module (ModuleType): The module to check for required arguments.
        args (Namespace): The arguments passed to the parser.
    """
    if hasattr(module, "add_parser"):
        module.add_parser(parser)

    if hasattr(module, "add_arguments"):
        module.add_arguments(parser)

def _add_optional_args_parser(parser: ArgumentParser, module: "ModuleType", args: Namespace):
    """
    Checks if the module's required arguments are present in the arguments
    passed to the parser. The module will be checked to see if it has all
    of the required arguments. If the module has all of the required
    arguments, then the module will be added to the parser.

    Args:
        parser (ArgumentParser): The parser to add the module to.
        module (ModuleType): The module to check for required arguments.
        args (Namespace): The arguments passed to the parser.
    """
    if not _has_required_args(module, args):
        return
    
    _add_args_parser(parser, module, args)

def initialize_parser(parser: ArgumentParser) -> ArgumentParser:
    """ 
    Add arguments to the parser. Loops through all modules in this package and
    calls the add_arguments function in each one if it exists and is callable.
    The modules are imported dynamically, so adding a new module to this package
    will automatically add its arguments to the parser as long as the module
    has an add_arguments function.

    While looping through each module in this package, we'll attempt to add
    any subparsers that are defined in the module. Modules which contain
    subparsers should define add_parser as a function which takes a parser as
    an argument.

    This function will skip the main and parser modules, as they are not
    intended to be used as subparsers or intended to add arguments to the
    parser.

    Args:
        parser (ArgumentParser): The parser to add arguments to.

    Returns:
        ArgumentParser: The parser with arguments added. Fluent interface.
    """
    PARENT_PACKAGE = __name__.rsplit(".", 1)[0]
    modules: list["ModuleType"] = [
        import_module(f".{module.removesuffix('.py')}", package=PARENT_PACKAGE)
        for module in listdir(dirname(__file__))
        if module.endswith(".py") and
        module not in ["main.py", "parser.py", "__init__.py"]
    ]

    modules_without_requirements = [
        module for module in modules if not hasattr(module, "ARGUMENTS")
    ]
    modules_with_requirements = [
        module for module in modules if hasattr(module, "ARGUMENTS")
    ]

    KNOWN_ARGS, _ = parser.parse_known_args()
    for module in modules_without_requirements:
        _add_args_parser(parser, module, KNOWN_ARGS)

    KNOWN_ARGS, _ = parser.parse_known_args() # More arguments may have been added
    for module in modules_with_requirements:
        _add_optional_args_parser(parser, module, KNOWN_ARGS)

    return parser

def get_arguments(parser: ArgumentParser) -> Namespace:
    """
    Get the arguments from the parser. If there is no input from stdin, then
    the arguments will be parsed from the command line. If there is input from
    stdin, then the arguments will be parsed from the command line and stdin after
    calling a flush on stdin.

    Args:
        parser (ArgumentParser): The parser to get arguments from.

    Returns:
        Namespace: The arguments from the parser.
    """
    arguments: list[str] = argv[1:]

    if not stdin.isatty():
        stdin.flush()
        arguments += stdin.readlines()

    return parser.parse_args(arguments)

__all__ = [
    "get_arguments",
]




# Copyright (c) 2023 Johnathan P. Irvin and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
from argparse import ArgumentParser, ArgumentTypeError

def limited_float(min: float, max: float):
    def _ret_func(arg: str):
        if not str(arg).isdecimal:
            raise ArgumentTypeError(f"{arg} is not a decimal")
        
        f = float(arg)
        if f < min or f > max:
            raise ArgumentTypeError(f"{arg} is not between {min} and {max}")
        
        return f
    
    return _ret_func


def add_arguments(parser: ArgumentParser):
    """
    Adds arguments that effect the results of the openai api call.

    '--count (-c)' is used to specify the number of results to return.
    '--temperature (-t)' is used to specify the sampling temperature to use.

    Args:
        parser (ArgumentParser): The parser to add the arguments to.
    """    
    parser.add_argument(
        "-c",
        "--count",
        type=int,
        default=1,
        help="The number of results to return.",
    )

    parser.add_argument(
        "-t",
        "--temperature",
        type=limited_float(0.0, 2.0),
        default=1.0,
        help="The sampling temperature to use. Must be between 0 and 2.",
    )


# Copyright (c) 2023 Johnathan P. Irvin and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
from argparse import ArgumentParser
from os import getppid

from psutil import NoSuchProcess, Process

import shell_craft.prompts as prompts


def get_calling_shell() -> str:
    """
    Get the name of the shell that called this script. This is used to determine
    the default prompt type. If the calling shell is PowerShell or pwsh, it is converted
    to "powershell" to match the prompt type. If the calling shell is not recognized,
    the default prompt type is "bash".

    If the calling shell cannot be determined, the default prompt type is "bash".

    Returns:
        str: The name of the shell.
    """
    try:
        return (
            "powershell" 
            if Process(getppid()).name() in 
            ["pwsh", "powershell"] else "bash"
        )
    except NoSuchProcess:
        return "bash"


def add_arguments(parser: ArgumentParser):
    """
    Adds '--prompt' as an argument to the parser. The choices are the names of
    the prompts in the prompts module, without the '_PROMPT' suffix, and converted to
    lowercase. The default is the name of the shell that called this script.

    Args:
        parser (ArgumentParser): The parser to add the argument to.
    """    
    parser.add_argument(
        "--prompt",
        type=str,
        choices=[
            prompt.removesuffix('_PROMPT').lower()
            for prompt in dir(prompts)
            if prompt.endswith("PROMPT")
        ],
        help="The type of prompt to use.",
        nargs='?',
        default=get_calling_shell()
    )



# Copyright (c) 2023 Johnathan P. Irvin and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
from shell_craft.prompts import Prompt


class PromptFactory:
    @staticmethod
    def get_prompt(prompt: str) -> Prompt:
        """
        Generates a new prompt object based on the prompt type.

        Args:
            prompt (str): A string representing the prompt type.

        Raises:
            ValueError: If the prompt type is not supported.

        Returns:
            Prompt: A new prompt object.
        """        
        import shell_craft.prompts as prompts

        available_prompts = [
            prompt.removesuffix('_PROMPT').casefold()
            for prompt in dir(prompts)
            if prompt.endswith("_PROMPT")
        ]

        if prompt.casefold() in available_prompts:
            return getattr(prompts, f"{prompt.upper()}_PROMPT")


        raise ValueError(f"Unknown prompt type: {prompt}")
